#!/usr/bin/env python

#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Tests for the outline.layer module.
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

# WARNING: Do not import builtins.str here as we do elsewhere in the code. Unit tests on Python 2
# need to preserve the existing Python 2 string type.
from builtins import range
import os
import sys
import unittest

import future.types
import mock

import outline
import outline.layer
import outline.io
import outline.modules.shell

from . import test_utils


SCRIPTS_DIR = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'scripts')


class CompositeTest(unittest.TestCase):
    """
    Tests for a composite layer. Composite layers
    can contain multiple layers.
    """

    def setUp(self):
        self.ol = outline.Outline()
        self.layer = outline.Layer("composite")
        self.ol.add_layer(self.layer)

        self.layer.add_child(outline.modules.shell.Shell("blah1", command=["ls", "-l"]))
        self.layer.add_child(outline.modules.shell.Shell("blah2", command=["ls", "-1"]))
        self.layer.add_child(outline.modules.shell.Shell("blah3", command=["ls"]))

        self.event = self.ol.get_layer("composite")

    @mock.patch('outline.Layer.system')
    def test_execute(self, systemMock):
        """Run the execute method."""
        with test_utils.TemporarySessionDirectory():
            self.ol.setup()
            self.event.execute(1)

        systemMock.assert_has_calls([
            mock.call(['ls', '-l'], frame=1),
            mock.call(['ls', '-1'], frame=1),
            mock.call(['ls'], frame=1),
        ])


class ChunkingTest(unittest.TestCase):

    """Tests layer chunking."""

    def setUp(self):
        self.ol = outline.load_outline(os.path.join(SCRIPTS_DIR, 'shell.outline'))
        self.ol.set_frame_range("1-10")

        self.event = self.ol.get_layer("cmd")
        self.event.set_chunk_size(5)

    def test_get_local_frame_set(self):
        """
        Test to make sure that that the localFrameSet is
        being constructed properly.  The local frame set
        is the frame list a particular frame is responsible
        for executing on the cue.  When the chunk-size is
        greated then 1, the local frame set will contain
        more than a single frame.
        """
        with test_utils.TemporarySessionDirectory():
            self.ol.setup()
        self.assertEqual([1, 2, 3, 4, 5], self.event.get_local_frame_set(1).getAll())
        self.assertEqual([8, 9, 10], self.event.get_local_frame_set(8).getAll())


class RangeTests(unittest.TestCase):

    """Tests that the proper frame ranges are being resolved."""

    def setUp(self):
        self.ol = outline.load_outline(os.path.join(SCRIPTS_DIR, 'shell.outline'))

    def test_no_layer_range_no_job_range(self):
        # No layer range, no outline range defaults to a single frame.
        self.assertEqual('1000-1000', self.ol.get_layer('cmd').get_frame_range())
        self.assertEqual(None, self.ol.get_frame_range())

    def test_no_layer_range_job_range(self):
        self.ol.set_frame_range('1000-2000')

        # No layer range, no outline range defaults to a single frame.
        self.assertEqual('1000-2000', self.ol.get_layer('cmd').get_frame_range())
        self.assertEqual('1000-2000', self.ol.get_frame_range())

    def test_layer_range_no_job_range(self):
        self.ol.get_layer('cmd').set_frame_range('1000-2000')
        self.assertEqual('1000-2000', self.ol.get_layer('cmd').get_frame_range())
        self.assertEqual(None, self.ol.get_frame_range())

    def test_layer_range_job_range(self):
        self.ol.set_frame_range('1000-2000')
        self.ol.get_layer('cmd').set_frame_range('1000-2000')

        expectedFrameStr = ','.join([str(i) for i in range(1000, 2001)])
        self.assertEqual(expectedFrameStr, self.ol.get_layer('cmd').get_frame_range())
        self.assertEqual('1000-2000', self.ol.get_frame_range())

    def test_intersecting_range(self):
        self.ol.set_frame_range('1000-2000x8')
        self.ol.get_layer('cmd').set_frame_range('1000-2000')

        expectedFrameStr = ','.join([str(i) for i in range(1000, 2001, 8)])
        self.assertEqual(expectedFrameStr, self.ol.get_layer('cmd').get_frame_range())
        self.assertEqual('1000-2000x8', self.ol.get_frame_range())

    def test_intersecting_failure(self):
        self.ol.set_frame_range('1000-1010')
        self.ol.get_layer('cmd').set_frame_range('1100-1200')

        self.assertFalse(self.ol.get_layer('cmd').get_frame_range())


class LayerTest(unittest.TestCase):
    """Tests for outline layer."""

    def setUp(self):
        """Setup a basis event from a preset outline file."""
        outline.config.add_section('Shell')
        outline.config.set('Shell', 'foo', 'bar')
        path = os.path.join(SCRIPTS_DIR, 'shell.outline')
        self.ol = outline.load_outline(path)
        self.ol.set_frame_range('1-10')
        self.ol.set_env('cue_test_01', 'foo')
        self.ol.set_env('cue_test_02', 'bar')
        self.layer = self.ol.get_layer('cmd')
        self.layer.set_env('cue_layer_01', 'layer-env-a')
        self.layer.set_env('cue_layer_02', 'layer-env-b')

    def tearDown(self):
        outline.config.remove_section('Shell')

    def test_name(self):
        """Test the name has been set properly."""
        self.assertEqual('cmd', self.layer.get_name())

    def test_to_string(self):
        """Tests to ensure __str__ returns the layer name."""
        self.assertEqual(self.layer.get_name(), str(self.layer))

    def test_get_set_args(self):
        """Test the argument getter/setter methods."""
        self.assertEqual(self.layer.get_arg('test1'), 1)
        self.assertEqual(self.layer.get_arg('test2'), 2)

        self.layer.set_arg('foo', 1)

        self.assertEqual(1, self.layer.get_arg('foo'))
        self.assertEqual(1, self.layer.get_arg('bar', 1))

        self.layer.set_default_arg('foo', 2)
        self.layer.set_default_arg('foo2', 2)
        self.assertEqual(1, self.layer.get_arg('foo'))
        self.assertEqual(2, self.layer.get_arg('foo2'))

    def test_invalid_type_args(self):
        """Test the interpolation of arg strings."""

        intArgName = 'some-int-arg'
        self.layer.require_arg(intArgName, int)
        self.assertRaises(
            outline.LayerException, self.layer.set_arg, intArgName, 'some-string-val')
        self.layer.set_arg(intArgName, 872)

        if sys.version_info[0] >= 3:
            strArgName = 'some-str-arg'
            self.layer.require_arg(strArgName, str)
            self.assertRaises(
                outline.LayerException, self.layer.set_arg, strArgName, {})
            self.layer.set_arg(strArgName, 'py3-string')
        else:
            strArgName = 'some-str-arg'
            self.layer.require_arg(strArgName, str)
            self.assertRaises(
                outline.LayerException, self.layer.set_arg, strArgName, {})
            self.layer.set_arg(strArgName, 'standard-py2-string')
            self.layer.set_arg(strArgName, 'py2-unicode')
            self.layer.set_arg(strArgName, future.types.newstr('py3-string-backport'))

            newstrArgName = 'some-newstr-arg'
            self.layer.require_arg(newstrArgName, future.types.newstr)
            self.assertRaises(
                outline.LayerException, self.layer.set_arg, newstrArgName, {})
            self.layer.set_arg(newstrArgName, 'standard-py2-string')
            self.layer.set_arg(newstrArgName, 'py2-unicode')
            self.layer.set_arg(newstrArgName, future.types.newstr('py3-string-backport'))

    def test_require_arg(self):
        """
        Test required arguments.  Unset required args will
        throw a LayerException if they are not set before
        setup() is run.
        """
        self.layer.require_arg('bobofet')
        self.assertRaises(outline.LayerException, self.layer.check_required_args)
        self.layer.set_arg('bobofet', 1)
        self.layer.check_required_args()

    def test_default_args(self):
        default_args = self.layer.get_default_args()

        self.assertEqual(
            {'chunk': 1, 'register': True, 'range': None, 'foo': 'bar'},
            default_args)

    def test_args_override(self):
        with test_utils.TemporarySessionDirectory():
            self.ol.setup()
            self.layer.put_data('args_override', {'arg_to_be_overridden': 'blah.blah'})

            self.layer.setup_args_override()

            self.assertEqual('blah.blah', self.layer.get_arg('arg_to_be_overridden'))

    def test_get_path(self):
        """Test that the layer session path is correct."""
        with test_utils.TemporarySessionDirectory():
            self.assertRaises(outline.OutlineException, self.layer.get_path)
            self.ol.setup()
            expectedPath = '%s/layers/%s' % (
                self.ol.get_session().get_path(), self.layer.get_name())
            self.assertEqual(expectedPath, self.layer.get_path())

    def test_setup(self):
        """Test setting up the event for launch."""
        with test_utils.TemporarySessionDirectory():
            self.layer.setup()

    @mock.patch('outline.Layer.system')
    def test_execute(self, systemMock):
        """Test execution of a frame."""
        os.environ = {}

        with test_utils.TemporarySessionDirectory():
            self.ol.setup()
            self.layer.execute(1)

            systemMock.assert_has_calls([mock.call(['ps', 'aux'], frame=1)])
            self.assertEqual('foo', os.environ['cue_test_01'])
            self.assertEqual('bar', os.environ['cue_test_02'])
            self.assertEqual('layer-env-a', os.environ['cue_layer_01'])
            self.assertEqual('layer-env-b', os.environ['cue_layer_02'])

    def test_get_set_frame_range(self):
        """Test getting/setting the frame range.  If the frame
        range is not set on a layer, then it should default to
        the outline frame range.
        """
        self.assertEqual(self.ol.get_frame_range(), self.layer.get_frame_range())
        self.layer.set_frame_range('1-10')
        self.assertEqual('1,2,3,4,5,6,7,8,9,10', self.layer.get_frame_range())

    def test_get_set_chunk_size(self):
        """Test get/set of chunk size."""
        self.layer.set_chunk_size(5)
        self.assertEqual(5, self.layer.get_chunk_size())

    def test_add_layer_during_setup(self):
        """Test to ensure that layers added during setup are serialized."""
        with test_utils.TemporarySessionDirectory():
            ol = outline.Outline('mr_hat')
            ol.add_layer(TestA('test_a'))
            ol.setup()

            ol_b = outline.load_outline(ol.get_session().get_file('outline.yaml'))
            # ensure the new layer was added
            self.assertTrue(ol_b.get_layer('test_b'))
            #  ensure that setup was actually run on the new layer
            self.assertTrue(ol_b.get_layer('test_b').is_setup)

    def test_after_init(self):
        ol = outline.Outline('after_init')
        ol.add_layer(TestAfterInit('test'))

        # Ensure after init was run,
        self.assertTrue(ol.get_layer('test').get_arg('after_init'))
        # Ensure that the layer has the right ol reference
        self.assertEqual(ol, ol.get_layer('test').get_outline())

    def test_after_init_current(self):
        ol = outline.Outline('after_init', current=True)
        TestAfterInit('test')

        # Ensure after init was run,
        self.assertTrue(ol.get_layer('test').get_arg('after_init'))
        # Ensure that the layer has the right ol reference
        self.assertEqual(ol, ol.get_layer('test').get_outline())

    def test_type_arg(self):
        """Test to ensure the type argument is handled properly."""
        outline.Outline.current = None
        t = TestA('test', type='Post')
        self.assertEqual('Post', t.get_type())

    def test_set_output_attribute(self):
        """Test setting an output attribute on all registered output."""
        t = TestA('test')
        t.add_output('node1', outline.io.Path('/tmp'))
        t.set_output_attribute('test', True)
        self.assertTrue(t.get_output('node1').get_attribute('test'))

    @mock.patch('outline.io.system')
    def test_system(self, systemMock):
        self.layer.system('arbitrary-command', ignore_error=True)

        systemMock.assert_called_with('arbitrary-command', True, None)

    def test_set_name(self):
        newLayerName = 'arbitrary-new-name'

        self.layer.set_name(newLayerName)

        self.assertEqual(newLayerName, self.layer.get_name())

        with test_utils.TemporarySessionDirectory():
            self.ol.setup()

            self.assertRaises(
                outline.LayerException, self.layer.set_name, 'this-should-fail')

    def test_should_fail_on_invalid_parent(self):
        invalid_object = 'some-string-object'

        self.assertRaises(outline.LayerException, self.layer.set_parent, invalid_object)

    def test_should_fail_on_invalid_child(self):
        invalid_object = 'some-string-object'

        self.assertRaises(outline.LayerException, self.layer.add_child, invalid_object)

    def test_should_add_event_listener(self):
        event_type = 'arbitrary-event-type'
        def callback(x):
            return x

        self.layer.add_event_listener(event_type, callback)

        self.assertEqual([callback], self.layer.get_event_handler().get_event_listeners(event_type))

    def test_should_fail_on_invalid_layer_type(self):
        invalid_layer_type = 'invalid-layer-type'

        self.assertRaises(outline.LayerException, self.layer.set_type, invalid_layer_type)

    def test_should_create_dependency_from_outline(self):
        with test_utils.TemporarySessionDirectory():
            outline.Outline.current = None
            ol = outline.Outline('after_init')
            ol.add_layer(TestAfterInit('test'))
            ol.add_layer(TestB('testb', require='test'))
            ol.setup()

            new_depend = ol.get_layer('testb').get_depends()[0]
            self.assertEqual('testb', new_depend.get_dependant_layer().get_name())
            self.assertEqual('test', new_depend.get_depend_on_layer().get_name())

    def test_should_create_dependency_from_layer(self):
        depend_on_layer = TestAfterInit('test')
        depend_er_layer = TestB('testb')

        depend_er_layer.depend_on(depend_on_layer)

        new_depend = depend_er_layer.get_depends()[0]
        self.assertEqual('testb', new_depend.get_dependant_layer().get_name())
        self.assertEqual('test', new_depend.get_depend_on_layer().get_name())

    def test_should_skip_duplicate_depend(self):
        depend_on_layer = TestAfterInit('test')
        depend_er_layer = TestB('testb')
        depend_er_layer.depend_on(depend_on_layer)
        num_depends = len(depend_er_layer.get_depends())

        depend_er_layer.depend_on(depend_on_layer)

        self.assertEqual(num_depends, len(depend_er_layer.get_depends()))

    def test_should_skip_depend_on_self(self):
        depend_layer = TestAfterInit('test')

        depend_layer.depend_on(depend_layer)

        self.assertEqual(0, len(depend_layer.get_depends()))

    def test_should_remove_depend(self):
        depend_on_layer = TestAfterInit('test')
        depend_er_layer = TestB('testb')
        depend_er_layer.depend_on(depend_on_layer)
        depend_to_remove = depend_er_layer.get_depends()[0]

        depend_er_layer.undepend(depend_to_remove)

        self.assertEqual(0, len(depend_er_layer.get_depends()))

    def test_should_get_dependents(self):
        with test_utils.TemporarySessionDirectory():
            depend_on_layer = TestAfterInit('test')
            depend_er_layer = TestB('testb', require='test')
            outline.Outline.current = None
            ol = outline.Outline('after_init')
            ol.add_layer(depend_on_layer)
            ol.add_layer(depend_er_layer)
            ol.setup()

            depend = depend_on_layer.get_dependents()[0]
            self.assertEqual('testb', depend.get_dependant_layer().get_name())
            self.assertEqual('test', depend.get_depend_on_layer().get_name())

    def test_should_add_inputs(self):
        layer = TestAfterInit('test')
        unnamed_input = outline.io.Path('/path/to/first/input')
        named_input = outline.io.Path('/path/to/second/input')
        bare_path = '/path/to/third/input'

        layer.add_input(None, unnamed_input)
        layer.add_input('named-input-name', named_input)
        layer.add_input('bare-path-name', bare_path)

        inputs = layer.get_inputs()
        self.assertEqual(unnamed_input, inputs['input0'])
        self.assertEqual(named_input, inputs['named-input-name'])
        self.assertEqual(bare_path, inputs['bare-path-name'].get_path())
        self.assertEqual(unnamed_input, layer.get_input('input0'))
        self.assertEqual(named_input, layer.get_input('named-input-name'))
        self.assertEqual(bare_path, layer.get_input('bare-path-name').get_path())

    def test_should_fail_on_duplicate_input(self):
        layer = TestAfterInit('test')
        input_to_add = outline.io.Path('/path/to/input')
        layer.add_input('input-name', input_to_add)

        self.assertRaises(outline.LayerException, layer.add_input, 'input-name', input_to_add)

    def test_should_check_input(self):
        layer = TestAfterInit('test')
        input_to_check = outline.io.Path('/path/to/input')
        input_to_check.set_attribute('checked', True)
        input_to_check.exists = mock.Mock()
        input_to_check.exists.return_value = True
        layer.add_input('input-name', input_to_check)

        layer.check_input()

        input_to_check.exists.assert_called()

    def test_should_skip_checking_input(self):
        layer = TestAfterInit('test')
        input_to_check = outline.io.Path('/path/to/input')
        input_to_check.set_attribute('checked', False)
        input_to_check.exists = mock.Mock()
        input_to_check.exists.return_value = True
        layer.add_input('input-name', input_to_check)

        layer.check_input()

        input_to_check.exists.assert_not_called()

    def test_should_fail_on_nonexistent_input(self):
        layer = TestAfterInit('test')
        input_to_check = outline.io.Path('/path/to/input')
        input_to_check.set_attribute('checked', True)
        input_to_check.exists = mock.Mock()
        input_to_check.exists.return_value = False
        layer.add_input('input-name', input_to_check)

        self.assertRaises(outline.LayerException, layer.check_input)

    def test_should_set_attribute_on_all_inputs(self):
        layer = TestAfterInit('test')
        input1 = outline.io.Path('/path/to/input')
        input2 = outline.io.Path('/path/to/another/input')
        layer.add_input('input-one', input1)
        layer.add_input('input-two', input2)
        attr_name = 'arbitrary-attribute-name'
        attr_val = 'arbitrary-attribute-value'

        layer.set_input_attribute(attr_name, attr_val)

        self.assertEqual(attr_val, input1.get_attribute(attr_name))
        self.assertEqual(attr_val, input2.get_attribute(attr_name))

    def test_should_add_outputs(self):
        layer = TestAfterInit('test')
        unnamed_output = outline.io.Path('/path/to/first/output')
        named_output = outline.io.Path('/path/to/second/output')
        bare_path = '/path/to/third/output'

        layer.add_output(None, unnamed_output)
        layer.add_output('named-output-name', named_output)
        layer.add_output('bare-path-name', bare_path)

        outputs = layer.get_outputs()
        self.assertEqual(unnamed_output, outputs['output0'])
        self.assertEqual(named_output, outputs['named-output-name'])
        self.assertEqual(bare_path, outputs['bare-path-name'].get_path())
        self.assertEqual(unnamed_output, layer.get_output('output0'))
        self.assertEqual(named_output, layer.get_output('named-output-name'))
        self.assertEqual(bare_path, layer.get_output('bare-path-name').get_path())

    def test_should_fail_on_duplicate_output(self):
        layer = TestAfterInit('test')
        output_to_add = outline.io.Path('/path/to/output')
        layer.add_output('output-name', output_to_add)

        self.assertRaises(outline.LayerException, layer.add_output, 'output-name', output_to_add)

    def test_should_check_output(self):
        layer = TestAfterInit('test')
        output_to_check = outline.io.Path('/path/to/output')
        output_to_check.set_attribute('checked', True)
        output_to_check.exists = mock.Mock()
        output_to_check.exists.return_value = True
        layer.add_output('output-name', output_to_check)

        layer.check_output()

        output_to_check.exists.assert_called()

    def test_should_skip_checking_output(self):
        layer = TestAfterInit('test')
        output_to_check = outline.io.Path('/path/to/output')
        output_to_check.set_attribute('checked', False)
        output_to_check.exists = mock.Mock()
        output_to_check.exists.return_value = True
        layer.add_output('output-name', output_to_check)

        layer.check_output()

        output_to_check.exists.assert_not_called()

    def test_should_fail_on_nonexistent_output(self):
        layer = TestAfterInit('test')
        output_to_check = outline.io.Path('/path/to/output')
        output_to_check.set_attribute('checked', True)
        output_to_check.exists = mock.Mock()
        output_to_check.exists.return_value = False
        layer.add_output('output-name', output_to_check)

        self.assertRaises(outline.LayerException, layer.check_output)

    def test_should_set_attribute_on_all_outputs(self):
        layer = TestAfterInit('test')
        output1 = outline.io.Path('/path/to/output')
        output2 = outline.io.Path('/path/to/another/output')
        layer.add_output('output-one', output1)
        layer.add_output('output-two', output2)
        attr_name = 'arbitrary-attribute-name'
        attr_val = 'arbitrary-attribute-value'

        layer.set_output_attribute(attr_name, attr_val)

        self.assertEqual(attr_val, output1.get_attribute(attr_name))
        self.assertEqual(attr_val, output1.get_attribute(attr_name))


class OutputRegistrationTest(unittest.TestCase):

    def setUp(self):
        outline.Outline.current = None

    def test_output_passing(self):
        """
        Test that output registered in a pre-process is serialized
        to a ol:outputs file in the render layer.
        """
        with test_utils.TemporarySessionDirectory():
            ol = outline.Outline("pre_test")

            # the render layer
            layer1 = TestA("test1")

            # the preprocess
            prelayer = outline.LayerPreProcess(layer1)
            prelayer._execute = lambda frames: prelayer.add_output(
                "test", outline.io.Path("/tmp/foo.#.exr"))

            # Add both to the outline
            ol.add_layer(layer1)
            ol.add_layer(prelayer)

            # setup for execute
            ol.setup()

            # now run the preprocess
            prelayer.execute(1000)

            # The file should exist.
            self.assertTrue(os.path.exists("%s/ol:outputs" % layer1.get_path()))

            # now run a single frame of the render layer and ensure that
            # the outputs are automatically loaded.
            layer1.execute(1000)
            self.assertEqual(1, len(layer1.get_outputs()))


class FrameTests(unittest.TestCase):

    def test_should_return_first_frame_of_outline_range(self):
        """Test getting/setting the frame range.  If the frame
        range is not set on a layer, then it should default to
        the outline frame range.
        """
        frame_range = '55-65'
        outline.Outline.current = None
        ol = outline.Outline('after_init')
        ol.set_frame_range(frame_range)
        layer = outline.layer.Frame('layer-name')
        ol.add_layer(layer)
        self.assertEqual('55', layer.get_frame_range())

    def test_should_return_default_frame_range(self):
        """Test getting/setting the frame range.  If the frame
        range is not set on a layer, then it should default to
        the outline frame range.
        """
        outline.Outline.current = None
        ol = outline.Outline('after_init')
        layer = outline.layer.Frame('layer-name')
        ol.add_layer(layer)
        self.assertEqual(outline.layer.DEFAULT_FRAME_RANGE, layer.get_frame_range())


class LayerPreProcessTests(unittest.TestCase):

    def test_should_be_util_type_and_preprocess_service(self):
        parent_layer_name = 'parent-layer'
        parent_layer = outline.layer.Layer(parent_layer_name)
        preprocess_layer = outline.layer.LayerPreProcess(parent_layer)

        self.assertEqual('Util', preprocess_layer.get_type())
        self.assertEqual('preprocess', preprocess_layer.get_service())

    def test_should_setup_depend_and_preprocess_layer(self):
        parent_layer_name = 'parent-layer'
        preprocess_layer_name = '%s_preprocess' % parent_layer_name
        parent_layer = outline.layer.Layer(parent_layer_name)
        preprocess_layer = outline.layer.LayerPreProcess(parent_layer)

        self.assertEqual(preprocess_layer_name, preprocess_layer.get_name())
        self.assertEqual(
            preprocess_layer_name, parent_layer.get_depends()[0].get_depend_on_layer().get_name())
        self.assertEqual(preprocess_layer, parent_layer.get_preprocess_layers()[0])

    def test_should_get_parent_layer(self):
        parent_layer_name = 'parent-layer'
        parent_layer = outline.layer.Layer(parent_layer_name)
        preprocess_layer = outline.layer.LayerPreProcess(parent_layer)

        self.assertEqual(parent_layer, preprocess_layer.get_creator())

    def test_should_get_first_frame_of_layer_range(self):
        parent_layer_name = 'parent-layer'
        parent_layer = outline.layer.Layer(parent_layer_name)
        parent_layer.set_frame_range('540-600')
        preprocess_layer = outline.layer.LayerPreProcess(parent_layer)

        self.assertEqual('540', preprocess_layer.get_frame_range())

    @mock.patch('outline.Layer.system', new=mock.Mock())
    def test_should_call_put_data_on_parent(self):
        os.environ = {}

        outline.Outline.current = None
        ol = outline.Outline('outline-name')
        parent_layer_name = 'parent-layer'
        parent_layer = outline.layer.Layer(parent_layer_name)
        parent_layer.put_data = mock.Mock()
        ol.add_layer(parent_layer)

        preprocess_layer = outline.layer.LayerPreProcess(parent_layer)
        output = outline.io.Path('/path/to/output')
        preprocess_layer.add_output('output-name', output)
        ol.add_layer(preprocess_layer)

        with test_utils.TemporarySessionDirectory():
            ol.setup()
            preprocess_layer.execute(1)

            parent_layer.put_data.assert_called_with(
                'ol:outputs', {'output-name': output}, force=True)


class LayerPostProcessTests(unittest.TestCase):

    def test_should_be_util_type(self):
        parent_layer_name = 'parent-layer'
        parent_layer = outline.layer.Layer(parent_layer_name)
        postprocess_layer = outline.layer.LayerPostProcess(parent_layer)

        self.assertEqual('Util', postprocess_layer.get_type())

    def test_should_setup_depend(self):
        parent_layer_name = 'parent-layer'
        postprocess_layer_name = '%s_postprocess' % parent_layer_name
        parent_layer = outline.layer.Layer(parent_layer_name)
        postprocess_layer = outline.layer.LayerPostProcess(parent_layer)

        self.assertEqual(postprocess_layer_name, postprocess_layer.get_name())
        self.assertEqual(
            parent_layer_name, postprocess_layer.get_depends()[0].get_depend_on_layer().get_name())

    def test_should_get_parent_layer(self):
        parent_layer_name = 'parent-layer'
        parent_layer = outline.layer.Layer(parent_layer_name)
        postprocess_layer = outline.layer.LayerPostProcess(parent_layer)

        self.assertEqual(parent_layer, postprocess_layer.get_creator())


class OutlinePostCommandTests(unittest.TestCase):

    def test_should_be_post_type_and_postprocess_service(self):
        parent_layer_name = 'parent-layer'
        parent_layer = outline.layer.Layer(parent_layer_name)
        post_layer = outline.layer.OutlinePostCommand(parent_layer)

        self.assertEqual('Post', post_layer.get_type())
        self.assertEqual('postprocess', post_layer.get_service())


class TestAfterInit(outline.Layer):
    __test__ = False
    def __init__(self, name, **args):
        outline.Layer.__init__(self, name, **args)

    def after_init(self, ol):
        self.set_arg("after_init", True)


class TestA(outline.Layer):
    __test__ = False
    def __init__(self, name, **args):
        outline.Layer.__init__(self, name, **args)

    def setup(self):
        self.get_outline().add_layer(TestB("test_b"))


class TestB(outline.Layer):
    __test__ = False
    def __init__(self, name, **args):
        outline.Layer.__init__(self, name, **args)
        self.is_setup = False

    def setup(self):
        self.is_setup = True


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Tests for the outline.backend.cue module.
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os
import unittest
import xml.etree.ElementTree as ET

import mock

import opencue_proto.job_pb2
import opencue.wrappers.job

import outline
import outline.backend.cue
import outline.cuerun
from .. import test_utils


SCRIPTS_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'scripts'))
TEST_USER = 'test-user'


class SerializeTest(unittest.TestCase):
    def testSerializeShellOutline(self):
        path = os.path.join(SCRIPTS_DIR, 'shell.outline')

        outline.config.set('outline', 'home', '/opencue/outline')
        outline.config.set('outline', 'user_dir', '/tmp/opencue/user')
        ol = outline.load_outline(path)
        launcher = outline.cuerun.OutlineLauncher(ol, user=TEST_USER)

        outlineXml = ET.fromstring(outline.backend.cue.serialize(launcher))

        self.assertEqual('spec', outlineXml.tag)
        self.assertEqual(1, len(outlineXml.findall('facility')))
        self.assertEqual('local', outlineXml.find('facility').text)
        self.assertEqual(1, len(outlineXml.findall('show')))
        self.assertEqual('testing', outlineXml.find('show').text)
        self.assertEqual(1, len(outlineXml.findall('shot')))
        self.assertEqual('default', outlineXml.find('shot').text)
        self.assertEqual(1, len(outlineXml.findall('user')))
        self.assertEqual(TEST_USER, outlineXml.find('user').text)
        self.assertEqual(1, len(outlineXml.findall('job')))
        job = outlineXml.find('job')
        self.assertEqual('shell', job.get('name'))
        self.assertEqual(1, len(job.findall('env')))
        self.assertEqual(0, len(list(job.find('env'))))
        self.assertEqual(1, len(job.findall('layers')))
        self.assertEqual(1, len(job.find('layers').findall('layer')))
        layer = job.find('layers').find('layer')
        self.assertEqual('cmd', layer.get('name'))
        self.assertEqual('Render', layer.get('type'))
        self.assertEqual(1, len(layer.findall('cmd')))
        self.assertEqual(
            '/opencue/outline/wrappers/opencue_wrap_frame '
            '/tmp/opencue/user '
            '/opencue/outline/bin/pycuerun '
            '{scripts_dir}/shell.outline '
            '-e #IFRAME#-cmd '
            '--version latest '
            '--repos  '
            '--debug'.format(scripts_dir=SCRIPTS_DIR), layer.find('cmd').text)
        self.assertEqual(1, len(layer.findall('range')))
        self.assertEqual('1000-1000', layer.find('range').text)
        self.assertEqual(1, len(layer.findall('chunk')))
        self.assertEqual('1', layer.find('chunk').text)
        self.assertEqual(1, len(layer.findall('services')))
        self.assertEqual(1, len(layer.find('services').findall('service')))
        self.assertEqual('shell', layer.find('services').find('service').text)
        self.assertEqual(1, len(outlineXml.findall('depends')))
        self.assertEqual(0, len(list(outlineXml.find('depends'))))


class CoresTest(unittest.TestCase):
    def setUp(self):
        # Ensure to reset current
        outline.Outline.current = None

    def create(self):
        ol = outline.Outline()
        layer = outline.Layer("test")
        ol.add_layer(layer)
        return ol, layer

    def assertCoresOverride(self, ol, v):
        launcher = outline.cuerun.OutlineLauncher(ol, user=TEST_USER)
        outlineXml = ET.fromstring(outline.backend.cue.serialize(launcher))
        job = outlineXml.find('job')
        layer = job.find('layers').find('layer')
        self.assertEqual(v, layer.find('cores').text)

    def testCores(self):
        ol, layer = self.create()
        layer.set_arg("cores", 42)
        self.assertCoresOverride(ol, "42.0")

    def testThreads(self):
        ol, layer = self.create()
        layer.set_arg("threads", 4)
        self.assertCoresOverride(ol, "4.0")

    def testCoresAndThreads(self):
        ol, layer = self.create()
        layer.set_arg("cores", 8)
        layer.set_arg("threads", 4)
        # cores overrides threads
        self.assertCoresOverride(ol, "8.0")

    def testNoCoreOverride(self):
        ol, layer = self.create()
        layer.set_arg("cores", None)

        launcher = outline.cuerun.OutlineLauncher(ol, user=TEST_USER)
        outlineXml = ET.fromstring(outline.backend.cue.serialize(launcher))
        job = outlineXml.find('job')
        layer = job.find('layers').find('layer')
        self.assertIsNone(layer.find('cores'))


class BuildCommandTest(unittest.TestCase):
    def setUp(self):
        path = os.path.join(SCRIPTS_DIR, 'shell.outline')
        outline.config.set('outline', 'home', '')
        outline.config.set('outline', 'user_dir', '')
        self.ol = outline.load_outline(path)
        self.launcher = outline.cuerun.OutlineLauncher(self.ol, user=TEST_USER)
        self.layer = self.ol.get_layer('cmd')

    def testBuildShellCommand(self):
        self.assertEqual(
            [
                '/wrappers/opencue_wrap_frame', '', '/bin/pycuerun',
                '%s/shell.outline -e #IFRAME#-cmd' % SCRIPTS_DIR,
                '--version latest', '--repos ', '--debug',
            ],
            outline.backend.cue.build_command(self.launcher, self.layer))

    def testBuildCommandWithStrace(self):
        self.layer.set_arg('strace', True)
        self.layer.set_arg('setshot', False)

        with test_utils.TemporarySessionDirectory():
            self.ol.setup()

            self.assertEqual(
                [
                    'strace', '-ttt', '-T', '-e', 'open,stat', '-f', '-o',
                    '%s/strace.log' % self.ol.get_session().get_path(self.layer),
                    '/wrappers/opencue_wrap_frame_no_ss', '', '/bin/pycuerun',
                    '%s -e #IFRAME#-cmd' % self.ol.get_path(),
                    '--version latest', '--repos ', '--debug',
                ],
                outline.backend.cue.build_command(self.launcher, self.layer))

    def testBuildCommandWithCustomWrapper(self):
        devUser = 'foo-user'
        wrapperPath = '/fake/wrapper'
        self.launcher.set_flag('dev', True)
        self.launcher.set_flag('devuser', devUser)
        self.layer.set_arg('wrapper', wrapperPath)

        self.assertEqual(
            [
                wrapperPath, '', '/bin/pycuerun',
                '%s/shell.outline -e #IFRAME#-cmd' % SCRIPTS_DIR,
                '--version latest', '--repos ', '--debug', '--dev',
                '--dev-user %s' % devUser,
            ],
            outline.backend.cue.build_command(self.launcher, self.layer))


class LaunchTest(unittest.TestCase):

    def setUp(self):
        self.job_wait_period_original = outline.backend.cue.JOB_WAIT_PERIOD_SEC
        outline.backend.cue.JOB_WAIT_PERIOD_SEC = .1

    def tearDown(self):
        outline.backend.cue.JOB_WAIT_PERIOD_SEC = self.job_wait_period_original

    @mock.patch('opencue.cuebot.Cuebot.getStub', new=mock.Mock())
    @mock.patch('opencue.Cuebot.setHosts')
    @mock.patch('opencue.api.launchSpecAndWait')
    def testLaunch(self, launchSpecAndWaitMock, setHostsMock):
        launchSpecAndWaitMock.return_value = [opencue.wrappers.job.Job()]
        serverName = 'foo-server'
        path = os.path.join(SCRIPTS_DIR, 'shell.outline')
        ol = outline.load_outline(path)
        launcher = outline.cuerun.OutlineLauncher(ol, user=TEST_USER)
        launcher.set_flag('server', serverName)
        serializedXml = launcher.serialize(use_pycuerun=True)

        outline.backend.cue.launch(launcher)

        launchSpecAndWaitMock.assert_called_with(serializedXml)
        setHostsMock.assert_called_with([serverName])

    @mock.patch('opencue.cuebot.Cuebot.getStub', new=mock.Mock())
    @mock.patch('opencue.api.isJobPending')
    @mock.patch('opencue.api.launchSpecAndWait')
    def testLaunchAndWait(self, launchSpecAndWaitMock, isJobPendingMock):
        jobName = 'some-job'
        launchSpecAndWaitMock.return_value = [
            opencue.wrappers.job.Job(opencue_proto.job_pb2.Job(name=jobName))]
        # Trigger one iteration of the wait loop.
        isJobPendingMock.side_effect = [True, False]
        path = os.path.join(SCRIPTS_DIR, 'shell.outline')
        ol = outline.load_outline(path)
        launcher = outline.cuerun.OutlineLauncher(ol, user=TEST_USER)
        launcher.set_flag('wait', True)
        serializedXml = launcher.serialize(use_pycuerun=True)

        outline.backend.cue.launch(launcher)

        launchSpecAndWaitMock.assert_called_with(serializedXml)
        isJobPendingMock.assert_has_calls([mock.call(jobName), mock.call(jobName)])

    @mock.patch('opencue.cuebot.Cuebot.getStub', new=mock.Mock())
    @mock.patch('opencue.api.getJob')
    @mock.patch('opencue.api.launchSpecAndWait')
    def testLaunchAndTest(self, launchSpecAndWaitMock, getJobMock):
        jobName = 'another-job'
        launchSpecAndWaitMock.return_value = [
            opencue.wrappers.job.Job(opencue_proto.job_pb2.Job(name=jobName))]
        getJobMock.return_value = opencue.wrappers.job.Job(
            opencue_proto.job_pb2.Job(name=jobName, state=opencue.api.job_pb2.FINISHED))

        path = os.path.join(SCRIPTS_DIR, 'shell.outline')
        ol = outline.load_outline(path)
        launcher = outline.cuerun.OutlineLauncher(ol, user=TEST_USER)
        launcher.set_flag('test', True)
        serializedXml = launcher.serialize(use_pycuerun=True)

        outline.backend.cue.launch(launcher)

        launchSpecAndWaitMock.assert_called_with(serializedXml)


if __name__ == '__main__':
    unittest.main()



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


"""
The outline event handler controls firing events to listeners.
"""


from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

from builtins import str
from builtins import object
import logging

from .exception import FailImmediately


logger = logging.getLogger("outline.event")

EVENT_TYPES = ("AFTER_INIT",
               "AFTER_PARENTED",
               "SETUP",
               "BEFORE_EXECUTE",
               "AFTER_EXECUTE",
               "AFTER_LAUNCH",
               "BEFORE_LAUNCH")

AFTER_INIT = EVENT_TYPES[0]
AFTER_PARENTED = EVENT_TYPES[1]
SETUP = EVENT_TYPES[2]
BEFORE_EXECUTE = EVENT_TYPES[3]
AFTER_EXECUTE = EVENT_TYPES[4]
AFTER_LAUNCH = EVENT_TYPES[5]
BEFORE_LAUNCH = EVENT_TYPES[6]


class EventHandler(object):
    """
    EventHandler keeps track of who is listening for which events.
    """
    def __init__(self, component):
        # pylint: disable=unused-private-member
        self.__component = component
        self.__listeners = {}

    def add_event_listener(self, event_type, callback):
        """
        Adds an event listener for the given type and
        callback function.
        """
        logger.debug("adding event listener %s", event_type)
        if event_type not in self.__listeners:
            self.__listeners[event_type] = []
        self.__listeners[event_type].append(callback)

    # pylint: disable=broad-except
    def emit(self, event):
        """Fires an event, calling any registered listeners."""
        logger.debug("fire event %s", event)
        for callback in self.__listeners.get(event.type, []):
            try:
                callback(event)
            except FailImmediately as fi:
                logger.debug("FailImmediately exception thrown, %s, %s", event.type, fi)
                raise fi
            except Exception as e:
                logger.debug("failed to execute event %s, %s", event.type, e)

    def get_event_listeners(self, event_type):
        """
        Return all the callback functions registered with
        a particular event type.
        """
        try:
            return self.__listeners[event_type]
        except KeyError:
            return []


class LaunchEvent(object):
    """
    A job launch event type.
    """
    def __init__(self, event_type, cuerun, **args):
        self.type = event_type
        self.cuerun = cuerun
        self.__dict__.update(args)

    def __str__(self):
        return str(self.__dict__)


class LayerEvent(object):
    """
    A LayerEvent occurs within a layer.
    """
    def __init__(self, event_type, layer, **args):
        self.type = event_type
        self.layer = layer
        self.__dict__.update(args)

    def __str__(self):
        return str(self.__dict__)



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


"""Outline configuration"""


from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import str

# pylint: disable=wrong-import-position
from future import standard_library
standard_library.install_aliases()
# pylint: enable=wrong-import-position

import getpass
import logging
import os
import pathlib
import platform
import tempfile
import configparser

__all__ = ['config', 'read_config_from_disk']
__file_path__ = pathlib.Path(__file__)

# Environment variables which can be used to define a custom config file.
__CONFIG_FILE_ENV_VARS = [
    # OUTLINE_CONFIG_FILE is the preferred setting to use.
    'OUTLINE_CONFIG_FILE',
    # OL_CONFIG is deprecated, but kept for now for backwards compatibility.
    'OL_CONFIG',
]


logger = logging.getLogger("outline.config")


def config_base_directory():
    """Returns the OpenCue config base directory.

    This platform-dependent directory, stored within your user profile, is used by
    OpenCue components as the default location for various configuration files. Typically
    if you store your config files in this location, there is no need to set environment
    variables to indicate where your config files are located -- OpenCue should recognize
    them automatically.

    NOTE: This work is ongoing. Over time more OpenCue components will start using this
    base directory. See https://github.com/AcademySoftwareFoundation/OpenCue/issues/785.

    :rtype: str
    :return: config file base directory
    """
    if platform.system() == 'Windows':
        return os.path.join(os.path.expandvars('%APPDATA%'), 'opencue')
    return os.path.join(os.path.expanduser('~'), '.config', 'opencue')


def read_config_from_disk():
    """Loads configuration settings from config file on the local system.

    The configuration file used is, in order of preference:
    - Path defined by the OUTLINE_CONFIG_FILE environment variable.
    - Path defined by the OL_CONFIG environment variable.
    - Path within the config base directory (i.e. ~/.config/opencue/outline.cfg)
    - The default outline.cfg file which is distributed with the outline library.

    :rtype: ConfigParser
    :return: config settings
    """
    pyoutline_root_dir = __file_path__.parent.parent
    default_user_dir = pathlib.Path(
        tempfile.gettempdir()) / 'opencue' / 'outline' / getpass.getuser()

    _config = configparser.ConfigParser()
    config_file = None

    for config_file_env_var in __CONFIG_FILE_ENV_VARS:
        logger.debug('Checking for outline config file path in %s', config_file_env_var)
        config_file_from_env = os.environ.get(config_file_env_var)
        if config_file_from_env and os.path.exists(config_file_from_env):
            config_file = config_file_from_env
            break

    if not config_file:
        config_from_user_profile = os.path.join(config_base_directory(), 'outline.cfg')
        logger.debug('Checking for outline config at %s', config_from_user_profile)
        if os.path.exists(config_from_user_profile):
            config_file = config_from_user_profile

    if not config_file:
        default_config_path = __file_path__.parent / 'outline.cfg'
        logger.info('Loading default outline config from %s', default_config_path)
        if default_config_path.exists():
            config_file = default_config_path

    if not config_file:
        raise FileNotFoundError('outline config file was not found')

    _config.read(str(config_file))

    # Add defaults to the config,if they were not specified.
    if not _config.get('outline', 'home'):
        _config.set('outline', 'home', str(pyoutline_root_dir))

    if not _config.get('outline', 'user_dir'):
        _config.set('outline', 'user_dir', str(default_user_dir))

    return _config


config = read_config_from_disk()



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Tests for the outline.json module.
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import os
import unittest
from xml.etree import ElementTree as Et

import mock

import outline
from . import test_utils


JSON_DIR = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'json')


class JsonTest(unittest.TestCase):

    def testJson(self):
        """Load in a json string"""
        s = ('{'
                 '"name": "test_job", '
                 '"range": "1-10", '
                 '"layers": [{'
                    '"name": "layer_1", '
                    '"module": "outline.modules.shell.Shell", '
                    '"env": {"LAYER_KEY1": "LAYER_VALUE1"}, '
                    '"command": ["/bin/ls"]'
                 '}]'
             '}')

        ol = outline.load_json(s)
        self.assertEqual('test_job', ol.get_name())
        self.assertEqual('1-10', ol.get_frame_range())
        self.assertEqual('LAYER_VALUE1', ol.get_layer('layer_1').get_env('LAYER_KEY1'))

        ol.get_layer('layer_1').set_env('LAYER_KEY2', 'LAYER_VALUE2')

        l = outline.cuerun.OutlineLauncher(ol)
        root = Et.fromstring(l.serialize())
        env1 = root.find('job/layers/layer/env/key[@name="LAYER_KEY1"]')
        self.assertEqual('LAYER_VALUE1', env1.text)
        env2 = root.find('job/layers/layer/env/key[@name="LAYER_KEY2"]')
        self.assertEqual('LAYER_VALUE2', env2.text)

    @mock.patch('outline.layer.Layer.system')
    @mock.patch.dict(os.environ, {}, clear=True)
    def testJsonFile(self, systemMock):
        """Load JSON from a file"""
        with open(os.path.join(JSON_DIR, 'shell.outline'), encoding='utf-8') as fp:
            ol = outline.load_json(fp.read())
        with test_utils.TemporarySessionDirectory():
            ol.setup()
            layer = ol.get_layer('shell_layer')
            self.assertEqual('LAYER_VALUE', layer.get_env('LAYER_KEY'))
            layer.execute('1000')

            systemMock.assert_has_calls([mock.call(['/bin/ls'], frame=1000)])
            self.assertEqual('LAYER_VALUE', os.environ['LAYER_KEY'])

    def testFacility(self):
        """Test facility from JSON"""
        with open(os.path.join(JSON_DIR, 'facility.json'), encoding='utf-8') as fp:
            ol = outline.load_json(fp.read())
            self.assertEqual('test_facility', ol.get_facility())


if __name__ == '__main__':
    unittest.main()



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Local backend module.

Runs the given job on the local machine, using a SQLite database to store state.

See outline.backend.__init__.py for a description of the PyOutline backend system.
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import object
import subprocess

import sqlite3

import FileSequence

import outline
import outline.versions


def build_command(ol, layer, frame):
    """
    Build and return a pycurun shell command for the given layer frame.

    :type  layer : Layer
    :param layer : The layer to build a command for.

    :rtype: string
    :return: The shell command to run for a the given layer.
    """
    command = []
    command.append("%s/local_wrap_frame" % outline.config.get("outline","wrapper_dir"))
    command.append(outline.config.get("outline", "user_dir"))
    command.append(ol.get_show())
    command.append(ol.get_shot())
    command.append("%s/pycuerun" % outline.config.get("outline", "bin_dir"))
    command.append("%s -e  %d-%s" % (ol.get_path(),  frame, layer.get_name()))
    command.append(" -v %s" % outline.versions.get_version("outline"))
    command.append(" -r %s" % outline.versions.get_repos())
    command.append("-D")

    return command


def launch(launcher, use_pycuerun=None):
    """
    Start the local dispatcher.
    """
    # pycuerun is not used in this backend, but we keep it as a parameter for compatibility
    # with other backends.
    del use_pycuerun

    dispatcher = serialize(launcher)
    dispatcher.dispatch()


def serialize(launcher):
    """
    Create a local dispatcher object.
    """
    return Dispatcher(launcher.get_outline())


def serialize_simple(launcher):
    """
    For local we can call the regular serialize function.
    """
    return serialize(launcher)


def build_frame_range(frame_range, chunk_size):
    """
    Return an array of frames with no duplicates and chunking applied.
    """
    frame_set = FileSequence.FrameSet(frame_range)
    frames = []
    if chunk_size > 1:
        if chunk_size >= len(frame_set):
            frames.append(frame_set[0])
        else:
            unique_frames = list(set(frame_set))
            for i, unique_frame in enumerate(unique_frames):
                if i % chunk_size == 0:
                    frames.append(unique_frame)
    else:
        frames = list(FileSequence.FrameSet(frame_range))
    return frames


class LocalFrameError(Exception):
    """
    Raised when a locally run frame has failed.
    """


class Dispatcher(object):
    """
    Local version of a job dispatcher, responsible for launching each frame, monitoring the
    result, and updating the local state.
    """

    def __init__(self, ol):
        self.__ol = ol

        self.__conn = sqlite3.connect(":memory:")
        self.__create_dispatch_list()

    def dispatch(self):
        """
        Run the frames of the job in sequence and record the result.
        """
        try:
            while True:
                l, f = self.__get_next_frame()
                if l is None and f is None:
                    break

                layer = self.__ol.get_layer(l)

                command = build_command(self.__ol, layer, f)
                try:
                    retcode = subprocess.call(command, shell=False)
                    if retcode != 0:
                        raise LocalFrameError("frame failed")
                except LocalFrameError:
                    # Failed to run frame
                    # Set frame to dead
                    c = self.__conn.cursor()
                    c.execute("UPDATE frames SET state=? WHERE layer=? AND frame=?",
                              ('DEAD', l, f))
                    self.__conn.commit()
        finally:
            print("Job is done")
            self.__conn.close()

    def __create_dispatch_list(self):
        """
        Creates a list of dispatchable frames.
        """
        c = self.__conn.cursor()
        c.execute(
            'create table frames '
            '(layer text, frame int, state string, layer_order int, frame_order int)')

        for layer in self.__ol.get_layers():

            frames = build_frame_range(layer.get_frame_range(),
                                       layer.get_chunk_size())
            for frame in frames:
                c.execute(
                    'insert into frames values (?,?,?,?,?)',
                    (layer.get_name(), frame, 'WAITING', self.__ol.get_layers().index(layer),
                     int(frame)))
        self.__conn.commit()

    def __get_next_frame(self):
        c = self.__conn.cursor()
        c.execute(
            "SELECT layer, frame, state FROM frames WHERE state='WAITING' "
            "ORDER BY frame_order,layer_order LIMIT 1")
        result = c.fetchone()

        if result:
            c.execute("UPDATE frames SET state=? WHERE layer=? AND frame=?",
                      ('RUNNING', result[0], result[1]))
            self.__conn.commit()
            return result[0], result[1]

        return None, None



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


"""Plugin for booking local cores after a job is launched."""


from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import str
import os
import logging
from socket import gethostname
import subprocess

import opencue
from outline import event


logger = logging.getLogger("outline.plugins.local")

USE_LOCAL_CORES = 1
USE_LOCAL_THREADS = 0


def init_cuerun_plugin(cuerun):
    """Initialize the cuerun plugin."""

    cuerun.get_parser().add_plugin_option(
        "-L", "--run-local",
        action="callback",
        type="int",
        nargs=1,
        help="Run local cores",
        callback=setup_event,
        callback_args=(cuerun,))

    cuerun.get_parser().add_plugin_option(
        "-T", "--run-local-threads",
        action="callback",
        type="int",
        nargs=1,
        help="Set number of threads for local cores to run per frame.",
        callback=setup_local_threads,
        callback_args=(cuerun,))


def setup_event(option, opt, value, parser, *args, **kwargs):
    """Main callback for the plugin, to set up the event listener."""
    del option
    del opt
    del parser
    del kwargs

    # Don't have a way to pass this to the event but
    # a global will do.
    # pylint: disable=global-statement
    global USE_LOCAL_CORES
    USE_LOCAL_CORES = value

    args[0].add_event_listener(event.BEFORE_LAUNCH, setup_local_cores)


def setup_local_threads(option, opt, value, parser, *args, **kwargs):
    """Callback for the --run-local-threads option."""

    del option
    del opt
    del parser
    del args
    del kwargs

    # pylint: disable=global-statement
    global USE_LOCAL_THREADS
    USE_LOCAL_THREADS = value


def deed_local_machine():
    """Deed the local machine to the current user."""

    user = os.environ.get("USER")
    show = opencue.api.findShow(os.environ.get("SHOW", "pipe"))
    try:
        owner = opencue.api.getOwner(user)
    except opencue.CueException:
        owner = show.createOwner(user)

    owner.takeOwnership(gethostname())


def setup_local_cores(e):
    """Configure the outline to use the local machine."""

    deed_local_machine()

    threads = USE_LOCAL_THREADS or USE_LOCAL_CORES
    outline = e.outline
    outline.set_arg("localbook", {"host": gethostname(),
                                  "cores": str(USE_LOCAL_CORES),
                                  "memory": get_half_host_memory(),
                                  "threads": str(threads),
                                  "gpu": str(0)})


def get_half_host_memory():
    """Returns half of the amount of RAM the local machine has."""

    with subprocess.Popen("vmstat -s", shell=True,
                          bufsize=1000, stdout=subprocess.PIPE).stdout as pipe:
        data = pipe.read().strip()
        data = int(data[0:data.find(" ")])
        data = data / 2
        return str(data)



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Module for setting up the PyOutline plugin system.
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import object
import logging
import sys

from outline.config import config


logger = logging.getLogger("outline.plugins")


class PluginManager(object):
    """
    Class responsible for instantiating the plugin system and loading desired plugins.
    """

    registered_plugins = []

    @classmethod
    def init_cuerun_plugins(cls, cuerun):
        """Initialize all registered plugins."""
        for plugin in cls.registered_plugins:
            try:
                plugin.init_cuerun_plugin(cuerun)
            except AttributeError:
                pass

    @classmethod
    def load_plugin(cls, module_name):
        """Load a single plugin and register it with the plugin manager."""
        logger.debug("importing [%s] outline plugin.", module_name)
        try:
            module = __import__(module_name,
                                globals(),
                                locals(),
                                [module_name])
            try:
                module.loaded()
            except AttributeError:
                pass
            cls.registered_plugins.append(module)

        except ImportError as e:
            sys.stderr.write("Warning: plugin load failed: %s\n" % e)

    @classmethod
    def init_plugin(cls, module_name, layer):
        """Initialize a plugin on the given layer."""
        try:
            logger.debug("importing [%s] outline plugin.", module_name)
            plugin = __import__(module_name, globals(), locals(), [module_name])
            try:
                plugin.init(layer)
            except AttributeError:
                pass
        except ImportError as e:
            sys.stderr.write("Warning: plugin load failed: %s\n" % e)

    @classmethod
    def load_all_plugins(cls):
        """Load and register all plugins listed in the config file."""
        def section_priority(section_needing_key):
            priority_option = "priority"
            if config.has_option(section_needing_key, priority_option):
                return config.getint(section_needing_key, priority_option)
            return 0

        sections = sorted(config.sections(), key=section_priority)

        for section in sections:
            if section.startswith("plugin:"):
                if config.getint(section, "enable"):
                    logger.debug("Loading plugin '%s'", section)
                    cls.load_plugin(config.get(section, "module"))

    @classmethod
    def get_plugins(cls):
        """Gets a list of all registered plugins."""
        return cls.registered_plugins



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


"""Global storage and data exchange."""


from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

from builtins import str
from builtins import object
import os
import logging
import shutil
import uuid
import yaml

import outline
import outline.exception
import outline.layer


__all__ = ["is_session_path",
           "Session"]

logger = logging.getLogger("outline.session")


def is_session_path(folder):
    """
    Return true if the specified folder contains an outline session.

    :type  folder: str
    :param folder: The folder to check.

    :rtype: boolean
    :return: true if the given folder contains an outline session.
    """
    if not folder:
        return False
    if not os.path.isdir(folder):
        folder = os.path.abspath(os.path.dirname(folder))
    if not os.path.exists(os.path.join(folder, 'session')):
        logger.info("script is not part of an existing session.")
        return False
    return True


class Session(object):
    """
    The session provides global storage space for an outline.  This
    global storage space is usually located in cue/cue_archive but
    may be in a different location based on L{outline.config}.

    Using put_file, get_file you can copy files into this location which
    can be used by all frames. Using put_data, get_data you can
    serialize and deserialize data into this session than can be used
    by all frames.
    """

    def __init__(self, ol):
        object.__init__(self)

        # The outline object that the session is for.
        self.__outline = ol

        # The unique session name is either set by creating a new
        # session or loading an existing one.
        self.__name = None

        # The path of a loaded session.
        self.__path = None

        if is_session_path(self.__outline.get_path()):
            self.__load_session()
        else:
            self.__create_session()

    def __create_session(self):
        """Create and set a new session path."""

        logger.info("Creating new session path.")
        try:
            job_name = "%s-%s-%s_%s" % (self.__outline.get_show(),
                                        self.__outline.get_shot(),
                                        self.__outline.get_user(),
                                        self.__outline.get_name())

            # This is unique session name.  Its a combination
            # of the job name and a random uuid.
            self.__name = "%s/%s" % (job_name, uuid.uuid1())

            # The base dir is where we copy the outline and
            # store session data.
            base_path = outline.config.get("outline", "session_dir")
            base_path = base_path.format(
                HOME=os.path.expanduser("~"),
                SHOW=self.__outline.get_show(),
                SHOT=self.__outline.get_shot())
            base_path = os.path.join(base_path, self.__name)

            # If the base dir doesn't exist, create it.  Be sure
            # to make a directory to store the layer data.
            if not os.path.exists(base_path):
                logger.info("creating session path: %s", base_path)
                old_mask = os.umask(0)
                try:
                    os.makedirs(base_path, 0o777)
                    os.mkdir("%s/layers" % base_path, 0o777)
                finally:
                    os.umask(old_mask)

            # finally, set the session path so any calls to get_path
            # will return the proper value.
            self.__path = base_path

        except OSError as exp:
            msg = "Failed to create session path: %s, reason: %s"
            raise outline.exception.SessionException(msg % (self.get_path(), exp))

    def __load_session(self):
        """Loads an existing session based on the current path."""

        session = "%s/session" % (os.path.dirname(self.__outline.get_path()))
        if not os.path.exists(session):
            msg = "Invalid session %s, session file does not exist."
            raise outline.exception.SessionException(msg % session)

        logger.info("loading session: %s", session)
        with open(session, encoding='utf-8') as file_object:
            try:
                data = yaml.load(file_object, Loader=yaml.Loader)
            except Exception as exp:
                msg = "failed to load session from %s, %s"
                raise outline.exception.SessionException(msg % (session, exp))
        self.__name = data
        self.__path = os.path.dirname(self.__outline.get_path())
        logger.info("session loaded: %s", self.__name)

    def get_name(self):
        """
        Return the name of the session.  The session name format is:
        I{non-unique id/unique uuid}

        :rtype: str
        :return: The name of the session.
        """
        return str(self.__name)

    def save(self):
        """Save the current session file."""

        with open(f"{self.get_path()}/session", "w", encoding="utf-8") as fp:
            fp.write(self.__name)

    @staticmethod
    def __layer_name(layer):
        """
        Helper function to conform an layer object or name
        into a string.

        :rtype: str
        :return: the name of the given layer.
        """
        if isinstance(layer, outline.layer.Layer):
            return layer.get_name()
        return str(layer)

    def sym_file(self, src, layer=None, rename=None):
        """
        Symlink a file into the session and return the path
        of the symlink. If the destination file already
        exists it will be deleted first.

        :type  src: str
        :param src: The source path for the file to symlink.

        :type  layer: L{Layer} or str
        :param layer: Layer to copy file into. [Optional]

        :type  rename: str
        :param rename: A new name for the file. [Optional]

        :rtype: str
        :return: The full path to the new file in the session.
        """
        dst = [self.get_path(layer)]
        if rename:
            dst.append(rename)
        else:
            dst.append(os.path.basename(src))

        dst_path = "/".join(dst)
        logger.info("creating session link %s", dst_path)
        try:
            os.unlink(dst_path)
        except (OSError, FileNotFoundError):
            pass
        os.symlink(os.path.abspath(src), dst_path)

        return dst_path

    def put_file(self, src, layer=None, rename=None):
        """
        Copy a file into the session and return the new path.  This
        is used to create local versions of input files for the
        job to use while its running.

        For example, if you had a scene file of some kind, you could:

           - I{session.put_file("comp/chambers/comp_v1.nuke",rename="scene")}

        Using, get_file, you could obtain the new full path to this file.

           - I{path = session.get_file("scene")}

        :type  src: str
        :param src: The source path for the file to copy.

        :type  layer: L{Layer} or str
        :param layer: Layer to copy file into. [Optional]

        :type  rename: str
        :param rename: A new name for the file. [Optional]

        :rtype: str
        :return: The full path to the new file in the session.
        """

        dst = [self.get_path(layer)]
        if rename:
            dst.append(rename)
        else:
            dst.append(os.path.basename(src))

        dst_path = "/".join(dst)

        logger.info("creating session file %s", dst_path)

        shutil.copy(os.path.abspath(src), dst_path)
        return dst_path

    def get_file(self, name, layer=None, check=True, new=False):
        """
        Retrieve the full path to the given file name previously stored
        with :meth:`put_file`.

        :type  name: str
        :param name: The unique identifier for the file.

        :type  layer: outline.layer.Layer or str
        :param layer: outline.layer.Layer or name of layer, optional

        :type check: bool
        :param check: If check is set, an exception is thrown if
                      the file does not exist, defaults to `True`

        :type new: bool
        :param new: If new is set and the file your getting already
                    exists, a SessionException is thrown.  This ensures
                    if your getting a new path to open in the session that
                    the file does not already exist. If new is specified,
                    check is automatically set to false, defaults to `False`

        :rtype: str
        :return: The full path to the file stored under the
                 given name.
        """
        # If new is activated then check should obviously be false.
        if new:
            check = False

        path = "%s/%s" % (self.get_path(layer), name)

        if check:
            if not os.path.exists(path):
                raise outline.exception.SessionException("The path %s does not exist." % path)

        if new:
            if os.path.exists(path):
                raise outline.exception.SessionException("The path %s already exists " % path)

        return path

    def put_data(self, name, data, layer=None, force=False):
        """
        Serialize a primitive variable or structure and store
        it into the session under the specified name.

        Call this to pass data from frame to frame.

        :type  name: str
        :param name: A unique name for the data
        :type  data: mixed
        :param data: Any python object that can be pickled.
        :type  layer: outline.layer.Layer or str
        :param layer: The layer to store the data under. Leave this
                      as None if the data is for the whole job.
                      [Optional]
        :type  force: bool
        :param force: Overwrite data with the same name if it exists.
                      [Optional]
        """

        path = "%s/%s" % (self.get_path(layer), name)
        if os.path.exists(path) and not force:
            raise outline.exception.SessionException("There is already data being \
                stored under this name.")

        with open(path, "w", encoding="utf-8") as fp:
            fp.write(yaml.dump(data))

    def get_data(self, name, layer=None):
        """
        Retrieve previously stored session data stored
        under the specified name.

        :type  name: str
        :param name: The name given to the data.
        :type  layer: L{Layer} or str
        :param layer: The layer the data was stored under. [Optional]

        :rtype: mixed
        :return: Previously stored data.
        """

        path = "%s/%s" % (self.get_path(layer), name)
        if not os.path.exists(path):
            raise outline.exception.SessionException("There is not data in the session \
                stored under that name.")

        logger.debug("opening data path for %s : %s", name, path)
        with open(path, encoding='utf-8') as file_object:
            try:
                return yaml.load(file_object, Loader=yaml.Loader)
            except Exception as exp:
                msg = "failed to load yaml data from %s, %s"
                raise outline.exception.SessionException(msg % (path, exp))

    def get_path(self, layer=None):
        """
        Return the session path for the current job.  If a layer is
        specified, will return the session path for the layer.

        The session path can be used for temporary file storage on
        the shot tree. See L{Layer} get_temp_dir() for a a way to obtain
        a temporary directory on the local hard drive.

        Note: get_path() creates the layer session paths on demand.
        So, you must use get_path(layer) to obtain the layer session path
        or it may not exist if you construct the path on your own.

        :type  layer: L{Layer} or str
        :param layer: Specify the layer to get its session path. [Optional]

        :rtype: str
        :return: The path where session data is stored.

        """
        if not self.__name:
            msg = "error, the session does not have a name."
            raise outline.exception.SessionException(msg)

        if not layer:
            return "%s" % self.__path

        layer_name = self.__layer_name(layer)
        path = "%s/layers/%s" % (self.__path,
                                 layer_name)
        if os.path.exists(path):
            return path
        old_mask = os.umask(0)
        try:
            os.mkdir(path, 0o777)
        except OSError:
            # Just eat this.  If it did actually fail
            # the whole process will fail pretty soon.
            pass
        finally:
            os.umask(old_mask)

        return path



#!/usr/bin/env python

#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Tests for the outline.executor module.
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import time
import unittest

import outline


class TaskExecutorTest(unittest.TestCase):

    def test_simple_threading(self):
        e = outline.TaskExecutor(5)
        e.execute(self.print_, "hello thread 1")
        e.execute(self.print_, "hello thread 2")
        e.execute(self.print_, "hello thread 3")
        e.execute(self.print_, "hello thread 4")
        e.execute(self.print_, "hello thread 5")
        e.wait()

    @staticmethod
    def print_(msg):
        print("Test Message: %s" % msg)
        time.sleep(1)


if __name__ == '__main__':
    unittest.main()



#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
OpenCue backend module.

Uses the OpenCue Python API to submit the given job to OpenCue for processing.

See outline.backend.__init__.py for a description of the PyOutline backend system.
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import str
import logging
import os
import sys
import time
from xml.dom.minidom import parseString
from xml.etree import ElementTree as Et

from packaging.version import Version

import FileSequence
import opencue

import outline
import outline.depend
import outline.exception
import outline.util
import outline.versions.main


logger = logging.getLogger("outline.backend.cue")

__all__ = ["launch",
           "serialize",
           "serialize_simple"]

JOB_WAIT_PERIOD_SEC = 5


def build_command(launcher, layer):
    """
    Build and return a pycuerun shell command for the given layer.

    :type  launcher : outline.cuerun.OutlineLauncher
    :param launcher : The outline launcher.

    :type  layer : Layer
    :param layer : The layer to build a command for.

    :rtype: list
    :return: The shell command to run for a the given layer.
    """
    command = []

    if layer.get_arg("strace"):
        command.append("strace")
        command.append("-ttt")
        command.append("-T")
        command.append("-e")
        command.append("open,stat")
        command.append("-f")
        command.append("-o")
        command.append("%s/strace.log" % layer.get_path())

    if layer.get_arg("wrapper"):
        wrapper = layer.get_arg("wrapper")
    elif layer.get_arg("setshot", True):
        wrapper = "%s/opencue_wrap_frame" % outline.config.get("outline", "wrapper_dir")
    else:
        wrapper = "%s/opencue_wrap_frame_no_ss" % outline.config.get(
            "outline", "wrapper_dir")

    command.append(wrapper)
    command.append(outline.config.get("outline", "user_dir"))
    command.append("%s/pycuerun" % outline.config.get("outline", "bin_dir"))
    command.append("%s -e #IFRAME#-%s" % (launcher.get_outline().get_path(),
                                          layer.get_name()))
    command.append("--version %s" % outline.versions.get_version("outline"))
    command.append("--repos %s" % outline.versions.get_repos())
    command.append("--debug")

    if launcher.get("dev"):
        command.append("--dev")

    if launcher.get("devuser"):
        command.append("--dev-user %s" % launcher.get("devuser"))

    return command


def launch(launcher, use_pycuerun=True):
    """
    Launch the given L{OutlineLauncher}.

    :type launcher: L{OutlineLauncher}
    :param launcher: The OutlineLauncher to launch.
    :type use_pycuerun: bool
    :param use_pycuerun: Enable/Disable pycuerun.

    :rtype: opencue.Entity.Job
    :return: The opencue job that was launched.
    """

    if launcher.get("server"):
        opencue.Cuebot.setHosts([launcher.get("server")])
        logger.info("cuebot host set to: %s", launcher.get("server"))

    jobs = opencue.api.launchSpecAndWait(launcher.serialize(use_pycuerun=use_pycuerun))

    if launcher.get("wait"):
        wait(jobs[0])
    elif launcher.get("test"):
        test(jobs[0])

    return jobs


def test(job):
    """
    Test the given job.  This function returns immediately
    when the given job completes, or throws an L{OutlineException}
    if the job fails in any way.

    :type job: opencue.Entity.Job
    :param job: The job to test.
    """
    logging.basicConfig(level=logging.DEBUG)
    logger.info("Entering test mode for job: %s", job.data.name)

    # Unpause the job.
    job.resume()

    try:
        while True:
            try:
                job = opencue.api.getJob(job.name())
                if job.data.job_stats.dead_frames + job.data.job_stats.eaten_frames > 0:
                    raise outline.exception.OutlineException(
                        "Job test failed, dead or eaten frames on: %s" % job.data.name)
                if job.data.state == opencue.api.job_pb2.FINISHED:
                    break
                logger.debug(
                    "waiting on %s job to complete: %d/%d", job.data.name,
                    job.data.job_stats.succeeded_frames, job.data.job_stats.total_frames)
            except opencue.CueException as ie:
                raise outline.exception.OutlineException(
                    "test for job %s failed: %s" % (job.data.name, ie))
            time.sleep(5)
    finally:
        job.kill()


def wait(job):
    """
    Wait for the given job to complete before returning.

    :type job: opencue.Entity.Job
    :param job: The job to wait on.
    """
    while True:
        try:
            if not opencue.api.isJobPending(job.data.name):
                break
            logger.debug(
                "waiting on %s job to complete: %d/%d", job.data.name,
                job.data.job_stats.succeeded_frames, job.data.job_stats.total_frames)
        except opencue.CueException as ie:
            print(
                "opencue error waiting on job: %s, %s. Will continue to wait." % (
                    job.data.name, ie),
                file=sys.stderr)
        time.sleep(JOB_WAIT_PERIOD_SEC)


def serialize(launcher):
    """
    Serialize the outline part of the given L{OutlineLauncher} into an OpenCue job specification,
    using pycuerun to wrap the job commands.

    :type launcher: L{OutlineLauncher}
    :param launcher: The outline launcher being used to launch the job.

    :rtype: str
    :return: A opencue job specification.
    """
    return _serialize(launcher, use_pycuerun=True)


def serialize_simple(launcher):
    """
    Serialize the outline part of the given L{OutlineLauncher} into an OpenCue job specification,
    skipping the pycuerun wrapper in favor of launching the job commands directly.

    :type launcher: L{OutlineLauncher}
    :param launcher: The outline launcher being used to launch the job.

    :rtype: str
    :return: A opencue job specification.
    """
    return _serialize(launcher, use_pycuerun=False)


def _warning_spec_version(spec_version, feature):
    logger.warning("spec_version=%s doesn't support %s", spec_version, feature)


def _serialize(launcher, use_pycuerun):
    """
    Serialize the outline part of the given L{OutlineLauncher} into a
    opencue job specification.

    :type launcher: L{OutlineLauncher}
    :param launcher: The outline launcher being used to launch the job.

    :rtype: str
    :return: A opencue job specification.
    """
    ol = launcher.get_outline()

    spec_version = Version(outline.config.get("outline", "spec_version"))

    root = Et.Element("spec")
    depends = Et.Element("depends")

    sub_element(root, "facility", launcher.get("facility"))
    sub_element(root, "show", launcher.get("show"))
    sub_element(root, "shot", launcher.get("shot"))
    user = launcher.get_flag("user")
    if not user:
        user = outline.util.get_user()
    sub_element(root, "user", user)
    if not launcher.get("nomail"):
        sub_element(root, "email", "%s@%s" % (user,
                                              outline.config.get("outline", "domain")))
    sub_element(root, "uid", str(outline.util.get_uid()))

    j = Et.SubElement(root, "job", {"name": ol.get_name()})
    sub_element(j, "paused", str(launcher.get("pause")))
    if spec_version >= Version("1.11"):
        sub_element(j, "priority", str(launcher.get("priority")))
    elif launcher.get("priority"):
        _warning_spec_version(spec_version, "priority")
    sub_element(j, "maxretries", str(launcher.get("maxretries")))
    if spec_version >= Version("1.13"):
        if ol.get_maxcores():
            sub_element(j, "maxcores", str(ol.get_maxcores()))
        if ol.get_maxgpus():
            sub_element(j, "maxgpus", str(ol.get_maxgpus()))
    else:
        if ol.get_maxcores():
            _warning_spec_version(spec_version, "maxcores")
        if ol.get_maxgpus():
            _warning_spec_version(spec_version, "maxgpus")
    sub_element(j, "autoeat", str(launcher.get("autoeat")))

    if ol.get_arg("localbook"):
        Et.SubElement(j, "localbook", ol.get_arg("localbook"))

    if launcher.get("os"):
        sub_element(j, "os", launcher.get("os"))
    elif os.environ.get("OL_OS", False):
        sub_element(j, "os", os.environ.get("OL_OS"))

    env = Et.SubElement(j, "env")
    for env_k, env_v in ol.get_env().items():
        # Only pre-setshot environment variables are
        # passed up to the cue.
        if env_v[1]:
            pair = Et.SubElement(env, "key", {"name": env_k})
            pair.text = env_v[0]

    layers = Et.SubElement(j, "layers")
    for layer in ol.get_layers():

        # Unregistered layers are in the job but don't show up on the cue.
        if not layer.get_arg("register"):
            continue

        # Don't register child layers with opencue.
        if layer.get_parent():
            continue

        # The layer will return a valid range if its range and
        # the job's range are compatible.  If not, skip launching
        # that layer.
        frame_range = layer.get_frame_range()
        if not frame_range:
            logger.info("Skipping layer %s, its range (%s) does not intersect "
                        "with ol range %s", layer, layer.get_arg("range"), ol.get_frame_range())
            continue

        spec_layer = Et.SubElement(layers, "layer",
                                   {"name": layer.get_name(),
                                    "type": layer.get_type()})
        if use_pycuerun:
            sub_element(spec_layer, "cmd",
                        " ".join(build_command(launcher, layer)))
        else:
            sub_element(spec_layer, "cmd", " ".join(layer.get_arg("command")))
        sub_element(spec_layer, "range", str(frame_range))
        sub_element(spec_layer, "chunk", str(layer.get_chunk_size()))

        # opencue specific options
        # Keeping 'threads' for backward compatibility
        cores = None
        if layer.is_arg_set("cores"):
            cores = layer.get_arg("cores")
        elif layer.is_arg_set("threads"):
            cores = layer.get_arg("threads")
        if cores is None:
            logger.debug("%s will use service cores.", layer.get_name())
        else:
            logger.debug("%s is set to override service cores.", layer.get_name())
            if layer.is_arg_set("cores") and layer.is_arg_set("threads"):
                logger.warning("%s has both cores and threads. Use cores.", layer.get_name())
            sub_element(spec_layer, "cores", "%0.1f" % float(cores))

        if layer.is_arg_set("threadable"):
            sub_element(spec_layer, "threadable",
                        bool_to_str(layer.get_arg("threadable")))

        if layer.get_arg("memory"):
            sub_element(spec_layer, "memory", "%s" % (layer.get_arg("memory")))

        gpus = None
        if layer.get_arg("gpus"):
            if spec_version >= Version("1.12"):
                gpus = layer.get_arg("gpus")
            else:
                _warning_spec_version(spec_version, "gpus")

        gpu_memory = None
        if layer.get_arg("gpu_memory"):
            if spec_version >= Version("1.12"):
                gpu_memory = layer.get_arg("gpu_memory")
            else:
                _warning_spec_version(spec_version, "gpu_memory")

        if gpus or gpu_memory:
            # Cuebot expects non-zero positive value on gpus and gpu_memory
            if gpus is None:
                gpus = 1
            if gpu_memory is None:
                gpu_memory = "1g"

            sub_element(spec_layer, "gpus", "%d" % gpus)
            sub_element(spec_layer, "gpu_memory", "%s" % gpu_memory)

        if layer.get_arg("timeout"):
            if spec_version >= Version("1.10"):
                sub_element(spec_layer, "timeout", "%s" % (layer.get_arg("timeout")))
            else:
                _warning_spec_version(spec_version, "timeout")

        if layer.get_arg("timeout_llu"):
            if spec_version >= Version("1.10"):
                sub_element(spec_layer, "timeout_llu", "%s" % (layer.get_arg("timeout_llu")))
            else:
                _warning_spec_version(spec_version, "timeout_llu")

        if os.environ.get("OL_TAG_OVERRIDE", False):
            sub_element(spec_layer, "tags",
                        scrub_tags(os.environ["OL_TAG_OVERRIDE"]))
        elif layer.get_arg("tags"):
            sub_element(spec_layer, "tags", scrub_tags(layer.get_arg("tags")))

        layer_limits = layer.get_limits()
        if layer_limits:
            limits = Et.SubElement(spec_layer, "limits")
            for limit_name in layer_limits:
                limit = Et.SubElement(limits, "limit")
                limit.text = limit_name

        layer_env = Et.SubElement(spec_layer, "env")
        for env_k, env_v in layer.get_envs().items():
            pair = Et.SubElement(layer_env, "key", {"name": "{}".format(env_k)})
            pair.text = env_v

        services = Et.SubElement(spec_layer, "services")
        service = Et.SubElement(services, "service")
        try:
            service.text = layer.get_service().split(",")[0].strip()
        except (AttributeError, IndexError):
            service.text = "default"

        if spec_version >= Version("1.15"):
            layer_outputs = Et.SubElement(spec_layer, "outputs")
            outputs = layer.get_outputs()
            for output_name in outputs:
                output_path = outputs[output_name]
                output = Et.SubElement(layer_outputs, "output", {"name": output_name})
                output.text = output_path.get_path()

        build_dependencies(ol, layer, depends)

    if not layers:
        raise outline.exception.OutlineException(
            "Failed to launch job. There are no layers with frame "
            "ranges that intersect the job's frame range: %s" % ol.get_frame_range())

    # Dependencies go after all of the layers
    root.append(depends)

    xml = [
        '<?xml version="1.0"?>',
        '<!DOCTYPE spec PUBLIC "SPI Cue  Specification Language" '
        '"http://localhost:8080/spcue/dtd/cjsl-%s.dtd">' % spec_version,
        Et.tostring(root).decode()
    ]

    result = "".join(xml)
    logger.debug(parseString(result).toprettyxml())
    return result


def scrub_tags(tags):
    """
    Ensure that layer tags pass in as a string are formatted properly.
    """
    if isinstance(tags, str):
        tags = [tag.strip() for tag in tags.split("|")
                if tag.strip().isalnum()]
    return " | ".join(tags)


def bool_to_str(value):
    """
    If the given value evaluates to True, return
    "True", else return "False"
    """
    if value:
        return "True"
    return "False"


def build_dependencies(ol, layer, all_depends):
    """
    Iterate through all the layer's dependencies and
    add them to the job spec.
    """
    for dep in layer.get_depends():

        depend = Et.SubElement(all_depends, "depend",
                               type=dep.get_type(),
                               anyframe=bool_to_str(dep.is_any_frame()))

        if dep.get_type() == outline.depend.DependType.LayerOnSimFrame:

            frame_range = dep.get_depend_on_layer().get_frame_range()
            first_frame = FileSequence.FrameSet(frame_range)[0]

            sub_element(depend, "depjob", ol.get_name())
            sub_element(depend, "deplayer", layer.get_name())
            sub_element(depend, "onjob", ol.get_name())
            sub_element(depend, "onframe", "%04d-%s"
                        % (first_frame, dep.get_depend_on_layer().get_name()))
        else:
            sub_element(depend, "depjob", ol.get_name())
            sub_element(depend, "deplayer", layer.get_name())
            sub_element(depend, "onjob", ol.get_name())
            sub_element(depend, "onlayer", dep.get_depend_on_layer().get_name())


def sub_element(root, tag, text):
    """Convenience method to create a sub element with text"""
    e = Et.SubElement(root, tag)
    e.text = text
    return e



#!/usr/bin/env python

#  Copyright Contributors to the OpenCue Project
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Tests for the outline.backend.local module.
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os
import unittest

import mock

import outline
import outline.backend.local
import outline.cuerun


SCRIPTS_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'scripts'))


class BuildCommandTest(unittest.TestCase):
    def setUp(self):
        path = os.path.join(SCRIPTS_DIR, 'shell.outline')
        outline.config.set('outline', 'home', '')
        outline.config.set('outline', 'user_dir', '')
        self.ol = outline.load_outline(path)
        self.launcher = outline.cuerun.OutlineLauncher(self.ol)
        self.layer = self.ol.get_layer('cmd')

    def testBuildShellCommand(self):
        frameNum = 47

        generatedCmd = outline.backend.local.build_command(self.ol, self.layer, frameNum)

        self.assertEqual(
            [
                '/wrappers/local_wrap_frame', '', 'testing', 'default', '/bin/pycuerun',
                '%s/shell.outline -e  %d-cmd' % (SCRIPTS_DIR, frameNum), ' -v latest',
                ' -r ', '-D'
            ], generatedCmd)


class SerializeTest(unittest.TestCase):
    def setUp(self):
        path = os.path.join(SCRIPTS_DIR, 'shell.outline')
        self.ol = outline.load_outline(path)
        self.launcher = outline.cuerun.OutlineLauncher(self.ol)

    def testSerialize(self):
        self.assertEqual(
            outline.backend.local.Dispatcher,
            outline.backend.local.serialize(self.launcher).__class__)

    def testSerializeSimple(self):
        self.assertEqual(
            outline.backend.local.Dispatcher,
            outline.backend.local.serialize_simple(self.launcher).__class__)


class BuildFrameRangeTest(unittest.TestCase):
    def testBuildFrameRange(self):
        self.assertEqual([3, 4, 5, 6, 7, 8, 9], outline.backend.local.build_frame_range('3-9', 1))

    def testBuildChunkedFrameRange(self):
        self.assertEqual([3, 7], outline.backend.local.build_frame_range('3-9', 4))

    def testBuildLargeChunkedFrameRange(self):
        self.assertEqual([3], outline.backend.local.build_frame_range('3-9', 87))


class DispatcherTest(unittest.TestCase):
    @mock.patch('subprocess.call')
    def testDispatch(self, subprocessCallMock):
        path = os.path.join(SCRIPTS_DIR, 'shell.outline')
        ol = outline.load_outline(path)
        launcher = outline.cuerun.OutlineLauncher(ol)
        subprocessCallMock.return_value = 0

        outline.backend.local.launch(launcher)


if __name__ == '__main__':
    unittest.main()
